/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string>

#include "DomainAccessControlWiiU.h"
#include "CString.h"
#include "KURL.h"
#include "PlatformAccounts.h"
#include "PlatformSSL.h"

#include <PlatformOS.h>
using namespace LibPlatform;

namespace WebCore
{
namespace DomainAccessControl 
{

std::list<DomainAccess> domainAccessList;
bool bAllowUnrestrictedDomains = false;

void AllowUnrestrictedDomains(){bAllowUnrestrictedDomains = true;}

void ClearDomainWhitelist()
{
	domainAccessList.clear();
}

bool AddDomainToWhitelist(const String &origin, bool subdomains)
{
	return AddDomainToWhitelist((DomainAccess){origin, subdomains});
}

bool AddDomainToWhitelist(const DomainAccess &domain)
{
	// if '*' is specified, then we don't care about any other domains, so clear out the list and add just the '*'
	if (domain.origin == "*")
	{
		if (bAllowUnrestrictedDomains)
		{
			OS::Report("Whitelisting '%s'\n", domain.origin.latin1().data());
			domainAccessList.clear();
			domainAccessList.push_back(domain);
			return true;
		}
		else
			return false;
	}

	// if '*' is already in the list, we don't care about adding any other domains
	const std::list<DomainAccess>::const_iterator i = domainAccessList.begin();
	if ((i != domainAccessList.end()) && (i->origin == "*"))
		return false;

	KURL url(ParsedURLString, domain.origin);

	if (!url.isValid())
		return false;

	String host = url.host();
	String protocol = url.protocol();

	if (host == "")
		return false;

	if (protocol != "http" && protocol != "https" && protocol != "ws" && protocol != "wss")
		return false;

	OS::Report("Whitelisting '%s' (subdomains: %s)\n", domain.origin.latin1().data(), domain.subdomains? "true" : "false");
	domainAccessList.push_back(domain);
	return true;

	/*
	1.
	 Let element be the access element to be processed. 

	2.
	 If the origin attribute is absent,
	 then this element is in error and the user agent must ignore this element. 

	3.
	 Let origin be the result of applying the rule for getting a single attribute value
	 to the value of the origin attribute.
	 If the result is a single U+002A ASTERISK (*) character,
	 then the user agent must prepend the U+002A ASTERISK to the access-request list and skip all steps below. 

	4.
	 If origin is not a valid IRI,
	 if it has components other than scheme and iauthority,
	 if it has no host component,
	 or if it has a iuser info component,
	 then this element is in error and the user agent must ignore this element. 

	5.
	 If the subdomains attribute is absent, then let sub domains be the value false.
	 Otherwise, or let sub domains be the result of applying the rule for getting a single attribute value to the value of the subdomains attribute. If the value of sub domains is not a valid boolean value, then let sub domains be the value false. 

	6.
	 Let scheme be the scheme component of origin. Let host be the host component of origin. Let port be the port component of origin. If there is no port component, the user agent must use the default value for the protocol that corresponds to scheme. 

	7.
	 If scheme is unsupported by the user agent,
	 then this element is in error and the user agent must ignore this element. 

	8.
	 If scheme is "http" or "https",
	 then the user agent must process the value of host using the ToASCII algorithm as per [RFC3490].

	9.
	 The user agent must append an item inside the access-request list that is the tuple:
		scheme, host, port, sub domains. 
	*/

	return true;
}

bool IsDomainAccessible(const String &origin, int *pError)
{
	KURL url(ParsedURLString, origin);

	if (pError)
		*pError = DOMAIN_ACCESS_OK;

#ifdef BUILD_FINAL_RETAIL
	if (url.protocol() == "http" && !LibPlatform::SSL::IsNonHTTPSCommunicationAllowed())
		return false;
#endif

	if (url.protocol() == "file")
		return true;

	if (LibPlatform::Accounts::GetServiceToken() == "")
	{
		if (pError)
			*pError = DOMAIN_ACCESS_NOT_ACQUIRED_AUTH_TOKEN;

		return false;
	}

	for (std::list<DomainAccess>::const_iterator i = domainAccessList.begin(); i != domainAccessList.end(); i++)
	{
		if (bAllowUnrestrictedDomains && i->origin == "*")
			return true;

		KURL url_iterator(ParsedURLString, i->origin);

		if (url.protocol() != url_iterator.protocol())
			continue;

		short port_iterator = url_iterator.hasPort()? url_iterator.port() : (url_iterator.protocol() == "https" || url_iterator.protocol() == "wss")? 443 : 80;
		short port = url.hasPort()? url.port() : (url.protocol() == "https" || url.protocol() == "wss")? 443 : 80;

		if (port_iterator != port)
			continue;

		if (!i->subdomains)
		{
			if (url_iterator.host() != url.host())
			{
				continue;
			}
		}
		else
		{
			if (url_iterator.host() != url.host())
			{
				size_t len_iterator = url_iterator.host().length();
				size_t len = url.host().length();

				if (len <= len_iterator)
					continue;

				// check if last len characters in url match url_iterator and that is preceded by '.'
				if (url.host().substring(len - len_iterator, len_iterator) != url_iterator.host())
					continue;

				if (url.host().substring(len - len_iterator - 1, 1) != ".")
					continue;
			}
		}

		return true;
	}

	if (pError)
		*pError = DOMAIN_ACCESS_NOT_WHITELISTED;

	return false;
}

}
}
