/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INPUT_H
#define INPUT_H

#include <string>
#include "float_vec.h"

// Callback type defines for KPAD.
typedef void (*RemoteControlDpdCallback )(signed chan, signed reason);
typedef void (*RemoteControlMplsCallback)(signed chan, signed reason);
typedef void (*RemoteExtensionCallback)(signed chan, signed reason);
typedef void (*RemoteSamplingCallback)(signed chan);
typedef void (*RemoteConnectCallback)(signed chan, signed reason);
typedef void (*KeyboardAttachCallback)(void *event);
typedef void (*KeyboardDetachCallback)(void *event);
typedef void (*KeyboardEventCallback)(void *event);
typedef void (*KeyboardTriggeredCallback)(void *event);
typedef void (*KeyboardPressedCallback)(void *event);
typedef void (*KeyboardReleasedCallback)(void *event);

namespace LibPlatform {
	static unsigned InputReadCount = 0;
	void ResetInputGlobals();

	class Input {
	private:
		// Singleton.
		Input(void);
		~Input(void);

	public:
		typedef enum {
			WII_REMOTE_MODE_LITE = 0,
			WII_REMOTE_MODE_NORMAL,
			WII_REMOTE_MODE_FULL,
		} RemoteMode;

        typedef enum {
            MIC_CLOSE_EVENT_FLAG = 1 << 0,
            MIC_EXIT_EVENT_FLAG = 1 << 1,
            MIC_OPEN_EVENT_FLAG = 1 << 2,
            MIC_READY_EVENT_FLAG = 1 << 3,
            MIC_CONNECTED_EVENT_FLAG = 1 << 4,
            MIC_DISCONNECTED_EVENT_FLAG = 1 << 5
        } eMIC_EVENT_FLAGS;

		static Input *GetInstance(void);
		static void DestroyInstance(void);

		void InitializeRemoteCallbacks(
			RemoteControlDpdCallback dpd_callback = 0,
			RemoteControlMplsCallback mpls_callback = 0,
			RemoteSamplingCallback sampling_callback = 0,
			RemoteExtensionCallback extension_callback = 0,
			RemoteConnectCallback connect_callback = 0
		);

		struct InputConfig {
			bool	m_bGamePadClick;
			bool	m_bGamePadScroll;
			bool	m_bGamePadNav; 
			bool	m_bRemoteClick;
			bool	m_bRemoteScroll;
			bool	m_bRemoteNav;
			bool	m_bSWKBDEnabled;
			bool	m_bUSBKBDEnabled;
			int 	m_iRemotesSupported;
		};

		/**
		@brief Access input config setup storage
		*/
		InputConfig *getInputConfig();
		InputConfig *getInputConfigDefaults();
		void SetInputConfig(const InputConfig &input_config);
		void SetInputConfigDefaults(const InputConfig &input_config);

		/**
		@brief Post to the system that application is about to return from background mode
		*/
		void BackFromBackground(void);

		/**
		@brief Initializes a KPAD controller. This must be called for each controller that you want to gather input from.
		@param channel Which controller to initialize.
		@param motion_plus Whether or not to read motion plus info from the Remote on the given channel.
		@param pos_play_radius
		@param pos_play_sensitivity 
		@param hori_play_radius
		@param hori_play_sensitivity
		@param dist_play_radius
		@param dist_play_sensitivity
		@param acc_play_radius
		@param acc_play_sensitivity 
		@param repeat_delay
		@param repeat_pulse
		*/
		void InitializeRemoteChannel(
			signed channel = 0,
			float pos_play_radius = 0.03f, 
			float pos_play_sensitivity = 0.5f,
			float hori_play_radius = 0.03f, 
			float hori_play_sensitivity = 1.0f,
			float dist_play_radius = 0.0f, 
			float dist_play_sensitivity = 1.0f,
			float acc_play_radius = 0.0f, 
			float acc_play_sensitivity = 0.1f,
			float repeat_delay = 0.75f,
			float repeat_pulse = 0.25f
		);

	#ifdef USE_AUTOMATION
		void ForceRemoteConnectCallback(unsigned channel);
		signed GetRemoteConnectStatus(unsigned channel);
		void SetRemoteConnectStatus(unsigned channel, signed newStatus);
		void QueueRemoteStatus(unsigned channel, void* remoteStatus, unsigned timestamp, signed delay);
		void ClearRemoteQueue(unsigned channel);
		void QueueGamePadStatus(unsigned channel, void* gamePadStatus, unsigned timestamp, signed delay);
		void ClearGamePadQueue(unsigned channel);
	#endif

		/**
		@brief Initializes a KPAD controller. This must be called for each controller that you want to gather input from.
		@param channel Which controller to disconnect.
		*/
		void DisconnectRemoteChannel(signed channel);

		/**
		@brief This must be called once if motion plus data is to be queried.
		*/
		void InitializeMotionPlus(void);

		void EnableMotionPlus(signed channel, bool enable);
		void EnableMotionPlus(signed channel, bool enable, unsigned char mode);
		void EnableDPD(signed channel, bool enable);
		bool IsDPDEnabled(signed channel);

		/**
		@param channel Which Remote to set.
		@param enable Enables/disables target Remote
		*/
		void SetRemoteEnabled(signed channel, bool enable);

		/**
		@param channel Which controller to check.
		@return True if a Remote is enabled on the given channel
		*/
		bool IsRemoteEnabled(signed channel);

		/**
		@param channel Which Remote extension to set.
		@param enable Enables/disables target Remote extension
		*/
		void SetExtensionEnabled(signed channel, bool enable);

		/**
		@param channel Which controller to check.
		@return True if a Remote extension is enabled on the given channel
		*/
		bool IsExtensionEnabled(signed channel);

		/**
		@param channel Which Remote to set.
		@param enable Enables/disables target GamePad
		*/
		void SetGamePadEnabled(signed channel, bool enable);

		/**
		@param channel Which controller to check.
		@return True if a GamePad is enabled on the given channel
		*/
		bool IsGamePadEnabled(signed channel);

		/**
		@param channel Which controller to check.
		@return True if a Remote is connected to the given channel
		*/
		bool HasRemoteController(signed channel);

		/**
		@param channel Which controller to check.
		@return True if the classic controller is connected to the Remote
		*/
		bool HasClassicController(signed channel);
		
		/**
		@param channel Which controller to check.
		@return True if the nunchuck controller is connected to the Remote
		*/
		bool HasNunChuckController(signed channel);

		/**
		@param channel Which controller to check.
		@return True if the GamePad is connected.
		*/
		bool HasGamePadController(signed channel);

		bool HasUSBKeyboard(signed channel);
		bool HasBalanceBoard(void);
		bool AlmostHasBalanceBoard(void);

		/**
		@brief Motion plus has to be manually reset when different extensions are plugged into the remote.
			Calling this will return the proper setting to Pass to KPADEnableMpls(channel, SETTING).
		@channel Which remote to check
		@return Which setting should be used when re-enabling the mpls
		*/
		unsigned char GetDesiredMotionPlusSetting(signed channel, unsigned char dev_type = 0);
		unsigned char GetCurrentMotionPlusSetting(signed channel);

		void SetMplsConfigFlag(signed channel, bool configComplete);
		bool GetMplsConfigFlag(signed channel);

		void SetExtensionConfigFlag(signed channel, bool configComplete);
		bool GetExtensionConfigFlag(signed channel);

		void SetExtensionStatus(signed channel, unsigned configComplete);
		bool RemoteHasMotionPlusInserted(signed channel);

		void StartMotionPlusCalibration(signed channel);
		void StopMotionPlusCalibration(signed channel);

		RemoteMode GetRemoteMode(signed channel);
		void SetRemoteMode(signed channel, RemoteMode newMode);

		bool IsExtensionReady(signed channel);

		/**
		@brief this returns a value between 0-1 while the remote is calibrating and negative at all other times
		*/
		float GetMotionPlusCalibrationStatus(signed channel);
		
		/**
		@brief This returns true if the calibration was stopped prematurely, then resets the value
		*/
		bool GetMotionPlusCalibrationFailed(signed channel);

		/**
		@brief Detect KPAD input on the given channel. 
			This should be called once before reading input data.
		@return Returns true if input is detected.
		*/
		bool ReadRemote(signed channel);
		void LockRemoteCache(void);
		void UnlockRemoteCache(void);
		unsigned HasCachedRemoteReads(signed channel);
		void ClearCachedRemoteReads(signed channel, unsigned clear_count);

		/**
		@brief Checks the pointer status.
		@param channel Which KPAD to check.
		@return Returns true if the recognition is reliable and in a valid position.
		*/
		bool CheckRemotePointerPosition(signed channel);
		bool CheckRemotePointerPosition(signed channel, signed buffer);

		/**
		@brief Returns the position of the pointer within the given projection space.
			Center is zero.
		@param channel Which KPAD channel to query.
		@param left Left most position of the screen-space projection.
		@param top Top most position of the screen-space projection.
		@param right Right most position of the screen-space projection.
		@param bottom Bottom most position of the screen-space projection.
		@return Returns the screen space position of the KPAD pointer.
		*/
		float2 GetRemotePointerPosition(signed channel, float screen_width, float screen_height);
		float2 GetRemotePointerPosition(signed channel, float screen_width, float screen_height, signed buffer);
		float2 GetRemotePointerPositionDiff(signed channel, float left, float top, float right, float bottom);
		float2 GetRemotePointerPositionOffset(signed channel, float screen_width, float screen_height, float width_offset, float height_offset);
		float GetRemotePointerPositionSpeed(signed channel);

		float GetRemoteDistance(signed channel);
		float GetRemoteDistance(signed channel, signed buffer);
		float GetRemoteDistanceDiff(signed channel);
		float GetRemoteDistanceSpeed(signed channel);

		float2 GetRemotePointerHorizon(signed channel);
		float2 GetRemotePointerHorizonDiff(signed channel);
		float GetRemotePointerHorizonSpeed(signed channel);

		void startRemoteVibrate(signed channel);
		void stopRemoteVibrate(signed channel);

		/**
		@brief Query channel for triggered input.
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key trigger.
		@return Returns true if any of the buttons on the given channel have been triggered.
		*/
		unsigned IsRemoteTriggered(signed channel);
		unsigned IsRemoteTriggered(signed channel, unsigned kpad_button);
		unsigned WasRemoteTriggered(signed channel, signed buffer);
		
		/**
		@brief Query channel for pressed input.
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key pressed.
		@return Returns true if any of the buttons on the given channel are currently pressed down.
		*/
		unsigned IsRemotePressed(signed channel);
		unsigned IsRemotePressed(signed channel, unsigned kpad_button);
		unsigned WasRemotePressed(signed channel, signed buffer);
		
		/**
		@brief Query channel for released input.
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key released.
		@return Returns true if any of the buttons on the given channel have been released.
		*/
		unsigned IsRemoteReleased(signed channel);
		unsigned IsRemoteReleased(signed channel, unsigned kpad_button);
		unsigned WasRemoteReleased(signed channel, signed buffer);

		/**
		@brief Query channel for battery level of remote.
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key released.
		*/
		unsigned GetRemoteBatteryLevel(signed channel);

		/**
		@brief Calculates the state Remoteas acceleration.
		@param channel Which controller to query.
		@return A 3D float3tor represention the acceleration.
		*/
		float3 GetRemoteAccelerationDirection(signed channel);
		float3 GetRemoteAccelerationDirection(signed channel, signed buffer);

		/**
		@brief The magnitude of the value returned from GetRemoteAccelerationDirection function.
		@param channel Which controller to query.
		@return Scaler (magnitude) of 3D acceleration.
		*/
		float GetRemoteAccelerationMagnitude(signed channel);
		float GetRemoteAccelerationMagnitude(signed channel, signed buffer);

		/**
		@brief Calculates the change in acceleration since last read.
		@param channel Which controller to query.
		@return A scaler representation of the change in magnitued.
		*/
		float GetRemoteAccelerationChange(signed channel);
		float GetRemoteAccelerationChange(signed channel, signed buffer);

		/**
		@brief Calculates the angular velocity. 360 dps is expressed as a value ranging from 0 to 1.
		@param channel Which controller to query
		@return A three dimensional float3tor representing the angular velocity.
		*/
		float3 GetRemoteAngularVelocity(signed channel);
		float3 GetRemoteAngularVelocity(signed channel, signed buffer);
		
		/**
		@brief Orientation of the Remote as a 3D float3tor.
		@param channel Which controller to listen to
		*/
		float3 GetRemoteAngle(signed channel);
		float3 GetRemoteAngle(signed channel, signed buffer);

		/**
		@brief Calulates the direction of the X axis of the Remote
		@param channel Which controller to listen to.
		*/
		float3 GetRemoteAxisDirectionX(signed channel);
		float3 GetRemoteAxisDirectionX(signed channel, signed buffer);
		
		/**
		@brief Calulates the direction of the Y axis of the Remote
		@param channel Which controller to listen to.
		*/
		float3 GetRemoteAxisDirectionY(signed channel);
		float3 GetRemoteAxisDirectionY(signed channel, signed buffer);

		/**
		@brief Calulates the direction of the Z axis of the Remote
		@param channel Which controller to listen to.
		*/
		float3 GetRemoteAxisDirectionZ(signed channel);
		float3 GetRemoteAxisDirectionZ(signed channel, signed buffer);

		/**
		@brief Query channel for triggered input from classic controller
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key trigger.
		@return Returns true if any of the buttons on the given channel have been triggered.
		*/
		unsigned IsClassicTriggered(signed channel);
		unsigned IsClassicTriggered(signed channel, unsigned kpad_button);
		unsigned WasClassicTriggered(signed channel, signed buffer);
		
		/**
		@brief Query channel for pressed input from classic controller.
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key pressed.
		@return Returns true if any of the buttons on the given channel are currently pressed down.
		*/
		unsigned IsClassicPressed(signed channel);
		unsigned IsClassicPressed(signed channel, unsigned kpad_button);
		unsigned WasClassicPressed(signed channel, signed buffer);
		
		/**
		@brief Query channel for released input from classic controller.
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key released.
		@return Returns true if any of the buttons on the given channel have been released.
		*/
		unsigned IsClassicReleased(signed channel);
		unsigned IsClassicReleased(signed channel, unsigned kpad_button);
		unsigned WasClassicReleased(signed channel, signed buffer);

		/**
		@brief Query channel for triggered input from NunChuck
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key trigger.
		@return Returns true if any of the buttons on the given channel have been triggered.
		*/
		unsigned IsNunChuckTriggered(signed channel);
		unsigned IsNunChuckTriggered(signed channel, unsigned kpad_button);
		unsigned WasNunChuckTriggered(signed channel, signed buffer);
		
		/**
		@brief Query channel for pressed input from NunChuck.
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key pressed.
		@return Returns true if any of the buttons on the given channel are currently pressed down.
		*/
		unsigned IsNunChuckPressed(signed channel);
		unsigned IsNunChuckPressed(signed channel, unsigned kpad_button);
		unsigned WasNunChuckPressed(signed channel, signed buffer);
		
		/**
		@brief Query channel for released input from NunChuck.
		@param channel Which KPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key released.
		@return Returns true if any of the buttons on the given channel have been released.
		*/
		unsigned IsNunChuckReleased(signed channel);
		unsigned IsNunChuckReleased(signed channel, unsigned kpad_button);
		unsigned WasNunChuckReleased(signed channel, signed buffer);

		/**
		@brief Calculates the state of the left analog stick of the classic controller.
		@param channel Which controller to query.
		@return A two dimensional float3tor representing the direction of the left stick.
		*/
		float2 GetClassicLStick(signed channel);
		float2 GetClassicLStick(signed channel, signed buffer);
		
		/**
		@brief Calculates the state of the right analog stick of the classic controller.
		@param channel Which controller to query.
		@return A two dimensional float3tor representing the direction of the right stick.
		*/
		float2 GetClassicRStick(signed channel);
		float2 GetClassicRStick(signed channel, signed buffer);

		/**
		@brief Calculates the state of the analog stick for the nunchuck.
		@param channel Wich controller to query.
		@return A two dimensional float3tor representing the direction of the nunchuck stick.
		*/
		float2 GetNunChuckStick(signed channel);
		float2 GetNunChuckStick(signed channel, signed buffer);

		/**
		@brief Calculates the state nunchuck's acceleration.
		@param channel Which controller to query.
		@return A 3D float3tor represention the acceleration.
		*/
		float3 GetNunChuckAccelerationDirection(signed channel);
		float3 GetNunChuckAccelerationDirection(signed channel, signed buffer);

		void GetVPADStatus(signed channel, void *curStatusP);	// just get a copy of the status

		/**
		@brief The magnitude of the value returned from GetNunChuckAccelerationDirection function.
		@param channel Which controller to query.
		@return Scaler (magnitude) of 3D acceleration.
		*/
		float GetNunChuckAccelerationMagnitude(signed channel);
		float GetNunChuckAccelerationMagnitude(signed channel, signed buffer);

		/**
		@brief Calculates the change in acceleration since last read.
		@param channel Which controller to query.
		@return A scaler representation of the change in magnitued.
		*/
		float GetNunChuckAccelerationChange(signed channel);
		float GetNunChuckAccelerationChange(signed channel, signed buffer);

		void InitializeGamePadChannel(
			signed channel = 0,
			float acc_play_radius = 0.0f, 
			float acc_play_sensitivity = 1.0f,
			float repeat_delay = 0.75f,
			float repeat_pulse = 0.25f
		);

		/**
		@brief Detect VPAD input on the given channel. 
			This should be called once before reading input data.
		@return Returns true if input is detected.
		*/
		bool ReadGamePad(signed channel);

		void LockGamePadCache(void);

		/**
		@brief Returns true past input status has been cached.
			Cached input is so that other threads, going at other speeds than the input loop
			can still gather input without falling behind the input thread.
		*/
		unsigned HasCachedGamePadReads(signed channel);

		/**
		@brief Removes all cached input data for the GamePad. 
		*/
		void ClearCachedGamePadReads(signed channel, unsigned clear_count);

		void UnlockGamePadCache(void);

		/**
		@brief If the gamepad has a valid touch position to be read,
			then this function will return true.
		*/
		bool CheckGamePadTouchPosition(signed channel);

		/**
		@brief Query VPAD touch screen for the most recent 'touch point'.
			This value gets updated every time ReadVPAD function is called on the given channel.
		@param channel Which VPAD controller to query.
		@return Returns the coordinates of the most current valid 'touch point'
		*/
		float2 GetGamePadTouchPosition(signed channel);

		/**
		@brief Query VPAD touch screen for the validity of the most recent 'touch point'.
		@param channel Which VPAD controller to query.
		@return Returns the validity of the most current 'touch point'
		*/
		unsigned GetGamePadTouchValidity(signed channel);

		/**
		@brief If the GamePad screen wasn't being touched last time ReadGamePad was called but it is now
			then this will return true.
		@param channel Which GamePad to listen to.
		*/
		bool IsGamePadTouchTriggered(signed channel);

		/**
		@brief If the GamePad screen was being touched the last two ReadGamePad calls then this will return true.
		@param channel Which GamePad to listen to.
		*/
		bool IsGamePadTouchPressed(signed channel);
		
		/**
		@brief If from the last two calls to ReadGamePad the screen was being touched in the first call, 
			but wasnt in the second this returns true.
		@param channel Which GamePad to listen to.
		*/
		bool IsGamePadTouchReleased(signed channel);

		float2 GetGamePadLStick(signed channel);
		float2 GetGamePadLStick(signed channel, signed buffer);
		float2 GetGamePadRStick(signed channel);
		float2 GetGamePadRStick(signed channel, signed buffer);

		float3 GetGamePadAccelerationDirection(signed channel);
		float3 GetGamePadAccelerationDirection(signed channel, signed buffer);
		float GetGamePadAccelerationMagnitude(signed channel);
		float GetGamePadAccelerationMagnitude(signed channel, signed buffer);
		float GetGamePadAccelerationChange(signed channel);
		float GetGamePadAccelerationChange(signed channel, signed buffer);

		void StartGamePadVibrate(signed channel, unsigned vibrationTime = 5000);
		void StopGamePadVibrate(signed channel);

		bool IsGamePadSensorBarEnabled(signed channel);
		void SetGamePadSensorBarEnabled(signed channel, bool enable);

		/**
		@brief Query channel for battery level of the GamePad.
		@param channel Which VPAD controller to query.
		*/
		unsigned char GetGamePadBatteryLevel(signed channel);

		/**
		@brief Calculates the angular velocity around the yaw, pitch, and roll axis of the GamePad.
		@param channel Which GamePad to read.
		@return A float3tor representing the angular velocity x y z - pitch yaw roll
		*/
		float3 GetGamePadGyro(signed channel);
		float3 GetGamePadGyro(signed channel, signed buffer);

		/**
		@brief Based off the gyro calculates the orientation of the GamePad..
		@param channel Which GamePad to read.
		@return A float3tor representing the orientation of the GamePad.
		*/
		float3 GetGamePadAngle(signed channel);
		float3 GetGamePadAngle(signed channe, signed buffer);

		/**
		@brief Calulates the direction of the X axis of the GamePad
		@param channel Which controller to listen to.
		*/
		float3 GetGamePadAxisDirectionX(signed channel);
		
		/**
		@brief Calulates the direction of the Y axis of the GamePad
		@param channel Which controller to listen to.
		*/
		float3 GetGamePadAxisDirectionY(signed channel);
		
		/**
		@brief Calulates the direction of the Z axis of the GamePad
		@param channel Which controller to listen to.
		*/
		float3 GetGamePadAxisDirectionZ(signed channel);

		/**
		@brief Query channel for triggered input.
		@param channel Which VPAD controller to query.
		@param vpad_button Which key to check. By default will check for any key pressed.
		@return Returns true if any of the buttons on the given channel are currently pressed down.
		*/
		unsigned IsGamePadTriggered(signed channel);
		unsigned IsGamePadTriggered(signed channel, unsigned vpad_button);
		unsigned WasGamePadTriggered(signed channel, signed buffer);
		
		/**
		@brief Query channel for pressed input.
		@param channel Which VPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key pressed.
		@return Returns true if any of the buttons on the given channel are currently pressed down.
		*/
		unsigned IsGamePadPressed(signed channel);
		unsigned IsGamePadPressed(signed channel, unsigned vpad_button);
		unsigned WasGamePadPressed(signed channel, signed buffer);
		
		/**
		@brief Query channel for released input.
		@param channel Which VPAD controller to query.
		@param kpad_button Which key to check. By default will check for any key released.
		@return Returns true if any of the buttons on the given channel have been released.
		*/
		unsigned IsGamePadReleased(signed channel);
		unsigned IsGamePadReleased(signed channel, unsigned vpad_button);
		unsigned WasGamePadReleased(signed channel, signed buffer);

		/**
		@brief Initializes library for detecting low level USB keyboard input.
		@param attach_callback Function that gets called when a keyboard is plugged in.
		@param dettach_callback Function t hat gets called when a keyboard is un-plugged.
		@param event_callback Function that is called when ever a key is pressed on the keyboard.
		*/
		void InitializeKeyboard(KeyboardAttachCallback attach_callback, KeyboardAttachCallback detach_callback, KeyboardEventCallback event_callback);

        /**
		@brief Query whether the last click was on the gamepad or not
		@param clickOnGamePad Whether or not the last click was from a GamePad
		@return Returns true if the last click was on a gamepad.  false if it was on the tv.
		*/
        void SetWasLastClickOnGamePad(bool clickOnGamePad);
        bool GetWasLastClickOnGamePad();

		/**
		@brief This must be called before any of the camera functions are used. 
			Calling this will allocate memory for the CAM library to work with.
		@param fps Frames per second for the camera stream. This can either be 15 or 30.
		*/
		void InitializeCamera(unsigned channel, unsigned fps);

		void DestroyCamera(unsigned channel);
		
		/**
		@brief YUV textures wont start streaming until this is called.
			Calling UpdateCameraStream when stream is closed will cause an error.
		*/
		void OpenCameraStream(unsigned channel);

		void PlayCameraStream(unsigned channel);

		/**
		@brief This should be called once a frame.
			Checks to see if the camera has finished encoding the image, and, if so, swapps the texture buffers.
		@return true if a new camera image has been sent from the GamePad.
		*/
		bool UpdateCameraStream(unsigned channel);

		void PauseCameraStream(unsigned channel);

		/**
		@brief Call to stop using the camera.
		*/
		void CloseCameraStream(unsigned channel);

		/**
		@brief Returns the most recently encoded image data (Y and UV textures)
		*/
		void *GetCameraImage(unsigned channel);

		bool CameraReady(unsigned channel);
		bool CameraStreamOpened(unsigned channel);
		bool CameraStreamClosed(unsigned channel);
		bool CameraExited(unsigned channel);

		bool CheckCameraReady(unsigned channel);

		/**
		@brief Returns the most recently encoded image data (Y and UV textures)
		*/
		float CameraGetFadeInAmount(unsigned channel);

		void BalanceBoardInitialize(void);
		void BalanceBoardZero(void);
		bool BalanceBoardRead(void);
		unsigned char BalanceBoardGetBatteryLevel(void);
		double BalanceBoardGetTGCWeight(void);
		double BalanceBoardGetRawWeight(unsigned index);
		double BalanceBoardGetAvgWeight(unsigned index);
        //Microphone functions

        int MicGetGamePadMicrophoneAudioBufferSize();
        void MicEventListenerRegistered(eMIC_EVENT_FLAGS flag); 
        void MicEventListenerUnregistered(eMIC_EVENT_FLAGS flag);
        void MicEventClearAllRegistrations();
    
        int MicInitGamePadMicrophone(unsigned channel, short *shortBufferPtr);
        int MicUninitGamePadMicrophone(unsigned channel);
        int MicOpenGamePadMicrophone(unsigned channel);
        int MicCloseGamePadMicrophone(unsigned channel);
        int MicSetEchoCancellation(unsigned channel, bool *enabled);
        int MicSetGainGamePadMicrophone(unsigned channel, float newGain, float *pOldGain);
        int MicGetGamePadMicrophoneGainValues(unsigned channel, float *pOutMinGain, float *pOutMaxGain, float *pOutCurrentGain);
        int MicUpdateGamePadMicrophone(unsigned channel);      
        bool MicIsGamePadMicrophoneOpen(unsigned channel);

        int TestMicReadyEvent(unsigned channel);
        int TestMicOpenEvent(unsigned channel);
        int TestMicCloseEvent(unsigned channel);
        int TestMicExitEvent(unsigned channel);
        int TestMicConnectedEvent(unsigned channel);
        int TestMicDisconnectedEvent(unsigned channel);        

	};
}

#endif//INPUT_H
